// starting code for Lab 12

// class for representing an 8-puzzle board state

import java.util.*;

public class Board {

    private int[] tiles;         // the contents of the board
    private int blankPos;        // the position of the blank
    private String action;       // the action that generated this board
    private Board parent;        // reference to parent board
    private int movesMade;       // number of moves made from start

    // constructor for the start state
    public Board(int[] tiles) {
        this.tiles = tiles;
        this.blankPos = find(0);
        this.action = null;
        this.parent = null;
        this.movesMade = 0;
    }

    // constructor for non-start states
    public Board(int[] tiles, String action, Board parent) {
        this.tiles = tiles;
        this.blankPos = find(0);
        this.action = action;
        this.parent = parent;
        this.movesMade = parent.movesMade + 1;
    }

    // returns the position number of the given tile
    private int find(int tile) {
        for (int i = 0; i < this.tiles.length; i++) {
            if (this.tiles[i] == tile) {
                return i;
            }
        }
        return -1;  // tile not found
    }

    // returns a printable representation of the board
    public String toString() {
        String s = "";
        for (int i = 0; i < this.tiles.length; i++) {
            if (i % 3 == 0) {
                s += "\n";
            }
            if (this.tiles[i] == 0) {
                s += "  ";
            } else {
                s += this.tiles[i] + " ";
            }
        }
        s += "\n";
        return s;
    }

    // returns the previous board state
    public Board getParent() {
        return this.parent;
    }

    // returns the action that produced this state from the parent
    public String getAction() {
        return this.action;
    }

    // returns the number of moves made from start to this state
    public int movesFromStart() {
        return this.movesMade;
    }

    // required by the List contains method
    public boolean equals(Object otherObject) {
        Board other = (Board) otherObject;
        for (int i = 0; i < this.tiles.length; i++) {
            if (this.tiles[i] != other.tiles[i]) {
                return false;
            }
        }
        return true;
    }

    // returns true if a tile can be moved in the specified direction
    public boolean canMoveTile(String direction) {
        if (direction.equals("up")) {
            // blank must not be in bottom row
            return !(this.blankPos == 6 || this.blankPos == 7 || this.blankPos == 8);
        } else if (direction.equals("down")) {
            // blank must not be in top row
            return !(this.blankPos == 0 || this.blankPos == 1 || this.blankPos == 2);
        } else if (direction.equals("left")) {
            // blank must not be in right column
            return !(this.blankPos == 2 || this.blankPos == 5 || this.blankPos == 8);
        } else if (direction.equals("right")) {
            // blank must not be in left column
            return !(this.blankPos == 0 || this.blankPos == 3 || this.blankPos == 6);
        } else {
            return false;
        }
    }

    // returns a new state resulting from moving a tile up
    public Board moveTileUp() {
        int tilePos = this.blankPos + 3;
        String action = String.format("slide %d up", this.tiles[tilePos]);
        return new Board(swapTiles(tilePos, this.blankPos), action, this);
    }

    // returns a new state resulting from moving a tile down
    public Board moveTileDown() {
        int tilePos = this.blankPos - 3;
        String action = String.format("slide %d down", this.tiles[tilePos]);
        return new Board(swapTiles(tilePos, this.blankPos), action, this);
    }

    // returns a new state resulting from moving a tile to the right
    public Board moveTileRight() {
        int tilePos = this.blankPos - 1;
        String action = String.format("slide %d to the right", this.tiles[tilePos]);
        return new Board(swapTiles(tilePos, this.blankPos), action, this);
    }

    // returns a new state resulting from moving a tile to the left
    public Board moveTileLeft() {
        int tilePos = this.blankPos + 1;
        String action = String.format("slide %d to the left", this.tiles[tilePos]);
        return new Board(swapTiles(tilePos, this.blankPos), action, this);
    }

    // returns a new tile array with tiles at positions i and j swapped
    private int[] swapTiles(int i, int j) {
        int[] copy = new int[this.tiles.length];
        for (int k = 0; k < this.tiles.length; k++) {
            copy[k] = this.tiles[k];
        }
        copy[i] = this.tiles[j];
        copy[j] = this.tiles[i];
        return copy;
    }

    //----------------------------------------------------------------------
    // methods to be completed...

    // returns a list of all new boards that can be generated in one move
    public List<Board> successors() {
        return null;  // temporary
    }
    
    // prints the action that generated this board from the parent board
    public void printAction() {
    }

    // prints the path from the start to this board
    public void printPath() {
    }

    public int compareTo(Board otherBoard) {
        return 0; // temporary
    }

    public int heuristic1(Board goal) {
        return 0; // temporary
    }

    public int heuristic2(Board goal) {
        return 0; // temporary
    }
    
    public void setPriority(Board goal) {
    }

    //----------------------------------------------------------------------
    // test boards

    public static Board goal1 = new Board(new int[] {1, 2, 3, 8, 0, 4, 7, 6, 5});
    public static Board a = new Board(new int[] {1, 3, 0, 8, 2, 4, 7, 6, 5});
    public static Board b = new Board(new int[] {1, 3, 4, 8, 6, 2, 0, 7, 5});
    public static Board c = new Board(new int[] {1, 3, 0, 4, 2, 5, 8, 7, 6});
    public static Board d = new Board(new int[] {8, 1, 2, 7, 0, 4, 6, 5, 3});
    public static Board e = new Board(new int[] {7, 1, 2, 8, 0, 3, 6, 5, 4});
    public static Board f = new Board(new int[] {2, 6, 3, 4, 0, 5, 1, 8, 7});
    public static Board g = new Board(new int[] {7, 3, 4, 6, 1, 5, 8, 0, 2});
    public static Board h = new Board(new int[] {7, 4, 5, 6, 0, 3, 8, 1, 2});
        
    // harder boards
    public static Board goal2 = new Board(new int[] {1, 2, 3, 4, 5, 6, 7, 8, 0});
    public static Board x = new Board(new int[] {5, 3, 4, 2, 7, 8, 6, 0, 1});
    public static Board y = new Board(new int[] {7, 5, 6, 4, 8, 0, 1, 2, 3});
    public static Board z = new Board(new int[] {3, 1, 7, 6, 4, 0, 8, 2, 5});

    //----------------------------------------------------------------------
    // demo program for Board class

    public static void main(String[] args) {
        Board board1 = Board.goal1;
        System.out.println(board1);
        Board board2 = board1.moveTileUp();
        board2.printAction();
        System.out.println(board2);
    }
}
